/**
Software package to simulate the Tumour--Immune System interaction model
presented in the paper 

"Tumour suppression by Immune System through Stochastic Oscillations"

by G.Caravagna, A.d’Onofrio, P.Milazzo and R.Barbuti.
 
Copyright (C) 2010,  Giulio Caravagna [caravagn@di.unipi.it]

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

package hsim;

import java.io.*;

public class AverageTestingUnit
{
    public static void main(String[] args) throws Exception 
    {
	
	/* PARAMETERS
	 * ----------
	 * antigenicity: the value of 'c'. 
	 * simulation_time: #days to simulate.
	 *
	 * This program computes the average of the eradication time for an instance of the model.
	 */
	double antigenicity = 71.5;
	int simulation_time = 2000;

	String [] param = {"-c", ""+antigenicity, "-TMAX", ""+simulation_time};
	Model m = new Model(param);
	
	
	// exec Model.STOCHASTIC_SAMPLE_SIZE simulations 
	int [] erad_table = new int[simulation_time];     // table of days of eradication
	int Not_erad = 0; // # of runs with no eradication
	
	
	System.out.println("Computing the probability density function for the eradication time of the following model.");
	System.out.println("\n\t"+m);
	System.out.println("Stochastic sample size: "+Model.STOCHASTIC_SAMPLE_SIZE);
		
	for(int i = 0; i < Model.STOCHASTIC_SAMPLE_SIZE; i++)
	    {
		m = new Model(param);
		
		// do a single run
		System.out.print("Simulation #"+i+" ... ");
		m.simulate();
		System.out.print("OK");
			
		// search for eradication instant
		int instant = (int) eradication(m);
		if(instant == -1) // not observed
		    {
			Not_erad++; // count it
			System.out.println(" [NOT OBSERVED]");
		    }
		else // here is observed
		    {
			erad_table[instant]++; // store the day in the table
			System.out.println(" [day #"+instant+"]");
		    }
		
	    } // process output
		
	System.out.println("\nALL Simulations completed ["+(Model.STOCHASTIC_SAMPLE_SIZE - Not_erad)+" eradications] ... Summary table");
		
	for(int j = 0; j < erad_table.length; j++)
	    {
		if(erad_table[j] != 0)
		    System.out.println("Day #"+j+" observed "+erad_table[j]+" eradications [probability "+((double)erad_table[j]/(Model.STOCHASTIC_SAMPLE_SIZE - Not_erad))+"%]");
	    }
	
	if(Not_erad != 0)
	    System.out.println("Probability of eradication = "+(Not_erad/Model.STOCHASTIC_SAMPLE_SIZE)+"% ["+(Model.STOCHASTIC_SAMPLE_SIZE - Not_erad)+" out of "+Model.STOCHASTIC_SAMPLE_SIZE+"]");
	else
	    System.out.println("Probability of eradication = 100% ["+(Model.STOCHASTIC_SAMPLE_SIZE - Not_erad)+" out of "+Model.STOCHASTIC_SAMPLE_SIZE+"]");
	
	// print output to file
	File fout = new File("density_"+antigenicity+"_"+simulation_time);
	FileWriter writer = new FileWriter(fout); 
	writer.write("# day \t probability\n");
	
	
	for(int i = 0; i < erad_table.length; i++)
	    {
		if(erad_table[i] != 0)
		    writer.write(i+"\t"+((double)erad_table[i]/(Model.STOCHASTIC_SAMPLE_SIZE - Not_erad))+"\n");
	    }
	
	writer.flush();
	writer.close();
    }


    // print output to file
    
    // search for eradication in the output of the tumor cells
    public static double eradication(Model m)
    {
	for(int i =0; i < m.outputT.length; i++)
	    {
		if(m.outputT[i] == 0.0) // found eradication
		    return m.outputTime[i]; // return time instant associated
	    }
	
	return -1;
    }
    
}