/**
Software package to simulate the Tumour--Immune System interaction model
presented in the paper 

"Tumour suppression by Immune System through Stochastic Oscillations"

by G.Caravagna, A.d’Onofrio, P.Milazzo and R.Barbuti.
 
Copyright (C) 2010,  Giulio Caravagna [caravagn@di.unipi.it]

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

package hsim;


public class Model 
{
	/********************************************************
	 * DEFAULT PARAMETERS		
	 */
	public static final double DEFAULT_T0 = 1.0;
	public static final double DEFAULT_E0 = 0.0;
	public static final double DEFAULT_IL0 = 0.0;

	public static final int STOCHASTIC_SAMPLE_SIZE = 2000;
	/********************************************************/

	/********************************************************
	 * MODEL CONFIGURATION		
	 */
	double [] state; // model state is {T, E, IL}
	
	double c = 0.0; // antigenicity (value to be specified ad each computation)
	double mu2 = 0.03;
	double p1 = 0.1245;
	double g1 = 2.0E7;
	double r2 = 0.18;
	double b = 1.0E-9;
	double a = 1;
	double g2 = 1.0E5;
	double p2 = 5;
	double g3 = 1.0E3;
	double mu3 = 10;
	double s1 = 0.0;
	double s2 = 0.0;
	double V = 3.2;
	
	double TMAX = 0.0; // simulation time (value to be specified ad each computation)

	// the output data structures
	int output_idx;
	double [] outputTime;
	double [] outputT;
	double [] outputE;
	double [] outputIL; 
	/********************************************************/

	public Model(String [] args)
	{
		try 
		{
			this.initparam(args);
		} 
		catch (Exception e) 
		{
			System.err.println("Error: initial configuration.\n\n"+e.getMessage());
			e.printStackTrace();
			System.exit(-1);
		}
	}
	
	// search for 'p' inside the array of strings 'args'
	// @return i>=0 if 'p' is found in 'args[i]', -1 otherwise
	public int search(String []args, String p)
	{
		for(int i = 0; i < args.length; i++)
		{
			if(args[i].compareTo(p) == 0)
				return i;
		}
		
		return -1;
	}
	
	public void initparam(String [] args)
	{
		// init default configuration
		state = new double[3];
		state[0] = Model.DEFAULT_T0;
		state[1] = Model.DEFAULT_E0;
		state[2] = Model.DEFAULT_IL0;
		
		// if required, change initial config.
		if(this.search(args,"-T0") != -1) // -T0 n0
			state[0] = Double.parseDouble(args[ this.search(args, "-T0") + 1]);
		
		if(this.search(args,"-E0") != -1) // -E0 n0
			state[1] = Double.parseDouble(args[ this.search(args, "-E0") + 1]);
		
		if(this.search(args,"-IL0") != -1) // -IL0 n0
			state[2] = Double.parseDouble(args[ this.search(args, "-IL0") + 1]);
		
		// look for the compulsory parameter 'antigenicity'
		if(this.search(args,"-c") != -1) // -c n0
			c = Double.parseDouble(args[ this.search(args, "-c") + 1]);
		else
		{
			System.err.println("Error: missing 'antigenicity' value.\n\n");
			System.exit(-1);
		}
		
		// look for the compulsory parameter 'simulation time'
		if(this.search(args,"-TMAX") != -1) // -c n0
			TMAX = Double.parseDouble(args[ this.search(args, "-TMAX") + 1]);
		else
		{
			System.err.println("Error: missing 'simulation time' value.\n\n");
			System.exit(-1);
		}
	}
	
	public void simulate()
	{
		// clean unusable objects by using the garbage collector
		System.gc();
		
		// output arrays
		output_idx = 0;
		outputTime = new double[(int) TMAX + 1];
		outputT = new double[(int)TMAX + 1];
		outputE = new double[(int)TMAX + 1];
		outputIL = new double[(int)TMAX + 1];
		
		// initialize a random number generator
		java.util.Random rgen = new java.util.Random(System.currentTimeMillis());
		
		double clock = 0.0;
		int oldclock = 0;
		
		// vectors for induced-changes by the firing of the actions
		int [] AGGT = { 1, -1, -1, 0, 0, 0, 0 };
		int [] AGGE = { 0, 0, 0, 1, -1, 1, 1 };
		
		double a1, a2, a3, a4, a5, a6, a7, SUMA;
		double Tn, En, In;
		double exp, Bn, told, tnew;
		double formula6, formula8, KTold, KDerivold;
		double [] vectA = new double[7];
		double choser, rand2;
		int reaz;
		
		int [] fired  = new int[7];
		
		while( clock < TMAX )
		{
			// at each step, fix the computation state
			Tn = state[0];
			En = state[1];
			In = state[2];
				
			// computes all the time--constant rate functions, and their summation
			a1 = r2*Tn;
			a2 = (b/V)*Tn*Tn*r2;
			a3 = (a*Tn*En)/(g2*V+Tn);
			a5 = mu2 * En;
			a6 = c*Tn;
			a7 = V*s1;
			SUMA = a1 + a2 + a3 + a5 + a6 + a7;
			
			// given an exponentially distributed number with mean 1
			exp = Math.log(1.0/rgen.nextDouble());
		    
			// compute the term 'Bn', formula (6)
			Bn = (1/mu3) * ( 
					( p2 / V ) * (Tn * En) * (1/(g3 * V + Tn) + s2) 
					); 
			
			// Newton-Raphson method used to solve the trascendental
			// equation for the putative time for next reaction
			told = -1;
			tnew = 0;
			
			while( Math.abs(tnew - told) > 0.0001 ) // 0.0001 min allowed step
			{
				told = tnew;
				
				// compute formula (8)
				formula8 = (p1 * En) / (( Bn + g1) * mu3);
				formula8 *= (told * mu3 * (Bn + g1) + g1  * 
						Math.log( (g1+In)/(-Bn + Math.exp(told * mu3) * (Bn + g1) + In ))
				);
							
				KTold = told * SUMA + formula8 - exp;
				KDerivold = SUMA + (En * p1 * In)/( g1 + In );
			
				tnew = told - (KTold / KDerivold);
			}// end: Newton-Raphson
			
			// update clock
			clock += tnew;
			
			// update number of interleukines (deterministically determined)
			formula6 = Bn + (In - Bn) * Math.exp(-mu3 * tnew);
			state[2] = formula6;
		
			// compute time--dependent propensity function 'a4(tnew-)'
			a4 = p1 * En * (formula6 / (g1 + formula6));
			SUMA += a4;
		
			// given all the propensities and their summation SUMA
			vectA[0] = a1;
			vectA[1] = a2;
			vectA[2] = a3;
			vectA[3] = a4;
			vectA[4] = a5;
			vectA[5] = a6;
			vectA[6] = a7;
			
			// choose the next stochastic events to fire
			rand2 = rgen.nextDouble();
		    
			reaz = 0;
			choser = rand2 * SUMA;
			
			for(reaz = 0; reaz < vectA.length; reaz++)
			{
				choser -= vectA[reaz];
				
			    	if( choser <= 0 )
					break;
			}

			if(reaz >= AGGT.length) return;
			
			fired[reaz]++;
				
			// apply the chosen event
			state[0] += AGGT[reaz];
			state[1] += AGGE[reaz];
			
			// prepare output
			if(oldclock < clock)
				{
					oldclock++;
					
					//System.out.print(clock+"\t"+state[0]+"\t"+state[1]+"\t"+state[2]+"\n");
					if(outputT.length > (int) Math.ceil(clock))
					{
					  // System.out.print((int)clock+" | ");
						
					    outputTime[output_idx] = clock;
						outputT[output_idx] = state[0];
						outputE[output_idx] = state[1];
						outputIL[output_idx] = state[2];
						output_idx++;
					}
				}
		}
		
		/** uncomment this for a summary table about the fired reactions 
		for(int i = 0; i < fired.length; i++)
			System.out.println("Reaction "+i+" fired "+fired[i]+" times.");
			
			*/
		trim_output();
	}
	
	// remove unused output space
	public void trim_output()
	{
		// use a buffer output space
		double [] buffer = new double [output_idx]; 
		
		// resize time
		System.arraycopy(outputTime, 0, buffer, 0, output_idx);
		outputTime = new double [output_idx];
		System.arraycopy(buffer, 0, outputTime, 0, output_idx);
		
		// resize T, E, IL
		System.arraycopy(outputT, 0, buffer, 0, output_idx);
		outputT = new double [output_idx];
		System.arraycopy(buffer, 0, outputT, 0, output_idx);
		
		System.arraycopy(outputE, 0, buffer, 0, output_idx);
		outputE = new double [output_idx];
		System.arraycopy(buffer, 0, outputE, 0, output_idx);
		
		System.arraycopy(outputIL, 0, buffer, 0, output_idx);
		outputIL = new double [output_idx];
		System.arraycopy(buffer, 0, outputIL, 0, output_idx);
	}
	
	public String toString()
	{
		return "Model: (T0,E0,IL0)=("+state[0]+
		", "+state[1]+", "+state[2]+") c="+c+" time="+TMAX+"\n";
	}
}
